@@ -37,6 +37,8 @@ module Sequel
     end
     
     class Dataset < Sequel::Dataset
+      SELECT_CLAUSE_METHODS = clause_methods(:select, %w'distinct columns from join where group having compounds order limit')
+      
       def fetch_rows(sql)
         execute(sql) do |result|
           begin
@@ -52,6 +54,12 @@ module Sequel
         end
         self
       end
+      
+      private
+      
+      def select_clause_methods
+        SELECT_CLAUSE_METHODS
+      end
     end
   end
 end
