@@ -116,7 +116,8 @@ end
 
 describe Sequel::Model do
   before do
-    @c = Class.new(Sequel::Model) do
+    @c = Class.new(Sequel::Model)
+    @c.class_eval do
       columns :score
       validates_each :score do |o, a, v|
         o.errors[a] << 'too low' if v < 87
@@ -551,12 +552,14 @@ end
 
 context "Superclass validations" do
   before do
-    @c1 = Class.new(Sequel::Model) do
+    @c1 = Class.new(Sequel::Model)
+    @c1.class_eval do
       columns :value
       validates_length_of :value, :minimum => 5
     end
     
-    @c2 = Class.new(@c1) do
+    @c2 = Class.new(@c1)
+    @c2.class_eval do
       columns :value
       validates_format_of :value, :with => /^[a-z]+$/
     end
@@ -601,7 +604,8 @@ end
 
 context ".validates with block" do
   specify "should support calling .each" do
-    @c = Class.new(Sequel::Model) do
+    @c = Class.new(Sequel::Model)
+    @c.class_eval do
       columns :vvv
       validates do
         each :vvv do |o, a, v|
@@ -938,14 +942,15 @@ end
 
 describe "Model#save" do
   before do
-    @c = Class.new(Sequel::Model(:people)) do
+    @c = Class.new(Sequel::Model(:people))
+    @c.class_eval do
       columns :id
 
       validates_each :id do |o, a, v|
         o.errors[a] << 'blah' unless v == 5
       end
     end
-    @m = @c.load(:id => 4)
+    @m = @c.load(:id => 4, :x=>6)
     MODEL_DB.reset
   end
 
@@ -958,14 +963,14 @@ describe "Model#save" do
     @m.id = 5
     @m.should be_valid
     @m.save.should_not be_false
-    MODEL_DB.sqls.should == ['UPDATE people SET id = 5 WHERE (id = 5)']
+    MODEL_DB.sqls.should == ['UPDATE people SET x = 6 WHERE (id = 5)']
   end
   
   specify "should skip validations if the :validate=>false option is used" do
     @m.raise_on_save_failure = false
     @m.should_not be_valid
     @m.save(:validate=>false)
-    MODEL_DB.sqls.should == ['UPDATE people SET id = 4 WHERE (id = 4)']
+    MODEL_DB.sqls.should == ['UPDATE people SET x = 6 WHERE (id = 4)']
   end
     
   specify "should raise error if validations fail and raise_on_save_faiure is true" do
