@@ -2,7 +2,7 @@ require File.join(File.dirname(__FILE__), 'spec_helper.rb')
 
 describe "Eagerly loading a tree structure" do
   before do
-    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.instance_variable_set(:@schemas, {})
     INTEGRATION_DB.create_table!(:nodes) do
       primary_key :id
       foreign_key :parent_id, :nodes
@@ -69,19 +69,12 @@ describe "Eagerly loading a tree structure" do
     clear_sqls
   end
   after do
-    Node.drop_table
+    INTEGRATION_DB.drop_table :nodes
     Object.send(:remove_const, :Node)
   end
 
   it "#descendants should get all descendants in one call" do
     nodes = Node.filter(:id=>1).eager(:descendants).all
-    sqls_should_be('SELECT * FROM nodes WHERE (id = 1)',
-      'SELECT * FROM nodes WHERE ((parent_id IN (1)) AND (id != parent_id))',
-      'SELECT * FROM nodes WHERE ((parent_id IN (2, 3)) AND (id != parent_id))',
-      'SELECT * FROM nodes WHERE ((parent_id IN (4)) AND (id != parent_id))',
-      'SELECT * FROM nodes WHERE ((parent_id IN (5)) AND (id != parent_id))',
-      'SELECT * FROM nodes WHERE ((parent_id IN (6)) AND (id != parent_id))',
-      'SELECT * FROM nodes WHERE ((parent_id IN (7)) AND (id != parent_id))')
     nodes.length.should == 1
     node = nodes.first
     node.pk.should == 1
@@ -104,17 +97,10 @@ describe "Eagerly loading a tree structure" do
     node.children.length.should == 1
     node.children.first.pk.should == 7
     node.children.first.parent.should == node
-    sqls_should_be
   end
 
   it "#ancestors should get all ancestors in one call" do
     nodes = Node.filter(:id=>[7,3]).order(:id).eager(:ancestors).all
-    sqls_should_be('SELECT * FROM nodes WHERE (id IN (7, 3)) ORDER BY id',
-      'SELECT * FROM nodes WHERE (id IN (1, 6))',
-      'SELECT * FROM nodes WHERE (id IN (5))',
-      'SELECT * FROM nodes WHERE (id IN (4))',
-      'SELECT * FROM nodes WHERE (id IN (2))',
-      'SELECT * FROM nodes WHERE (id IN (1))')
     nodes.length.should == 2
     nodes.collect{|x| x.pk}.should == [3, 7]
     nodes.first.parent.pk.should == 1
@@ -130,7 +116,6 @@ describe "Eagerly loading a tree structure" do
     node = node.parent
     node.parent.pk.should == 1
     node.parent.parent.should == nil
-    sqls_should_be
   end
 end
 
@@ -138,31 +123,23 @@ describe "Association Extensions" do
   before do
     module ::FindOrCreate
       def find_or_create(vals)
-        # Exploits the fact that Sequel filters are ruby objects that
-        # can be introspected.
-        author_id = @opts[:where].args[1]
-        first(vals) || \
-          @opts[:models][nil].create(vals.merge(:author_id=>author_id))
+        first(vals) || model.create(vals.merge(:author_id=>model_object.pk))
       end 
+      def find_or_create_by_name(name)
+        first(:name=>name) || model.create(:name=>name, :author_id=>model_object.pk)
+      end
     end
-    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.instance_variable_set(:@schemas, {})
     INTEGRATION_DB.create_table!(:authors) do
       primary_key :id
     end
     class ::Author < Sequel::Model
-      one_to_many :authorships, :extend=>FindOrCreate, :dataset=>(proc do
-        key = pk
-        ds = Authorship.filter(:author_id=>key)
-        ds.meta_def(:find_or_create_by_name) do |name|
-          first(:name=>name) || Authorship.create(:name=>name, :author_id=>key)
-        end 
-        ds  
-      end)
+      one_to_many :authorships, :extend=>FindOrCreate
     end
     INTEGRATION_DB.create_table!(:authorships) do
       primary_key :id
       foreign_key :author_id, :authors
-      text :name
+      String :name
     end
     class ::Authorship < Sequel::Model
       many_to_one :author
@@ -171,46 +148,32 @@ describe "Association Extensions" do
     clear_sqls
   end
   after do
-    Authorship.drop_table
-    Author.drop_table
+    INTEGRATION_DB.drop_table :authorships, :authors
     Object.send(:remove_const, :Author)
     Object.send(:remove_const, :Authorship)
   end
 
   it "should allow methods to be called on the dataset method" do
     Authorship.count.should == 0
-    sqls_should_be('SELECT COUNT(*) FROM authorships LIMIT 1')
     authorship = @author.authorships_dataset.find_or_create_by_name('Bob')
-    sqls_should_be("SELECT * FROM authorships WHERE ((author_id = 1) AND (name = 'Bob')) LIMIT 1",
-      /INSERT INTO authorships \((author_id, name|name, author_id)\) VALUES \((1, 'Bob'|'Bob', 1)\)/,
-      "SELECT * FROM authorships WHERE (id = 1) LIMIT 1")
     Authorship.count.should == 1
     Authorship.first.should == authorship
-    sqls_should_be('SELECT COUNT(*) FROM authorships LIMIT 1', "SELECT * FROM authorships LIMIT 1")
     authorship.name.should == 'Bob'
     authorship.author_id.should == @author.id
     @author.authorships_dataset.find_or_create_by_name('Bob').should == authorship
-    sqls_should_be("SELECT * FROM authorships WHERE ((author_id = 1) AND (name = 'Bob')) LIMIT 1")
     Authorship.count.should == 1
-    sqls_should_be('SELECT COUNT(*) FROM authorships LIMIT 1')
     authorship2 = @author.authorships_dataset.find_or_create(:name=>'Jim')
-    sqls_should_be("SELECT * FROM authorships WHERE ((author_id = 1) AND (name = 'Jim')) LIMIT 1",
-      /INSERT INTO authorships \((author_id, name|name, author_id)\) VALUES \((1, 'Jim'|'Jim', 1)\)/,
-      "SELECT * FROM authorships WHERE (id = 2) LIMIT 1")
     Authorship.count.should == 2
-    sqls_should_be('SELECT COUNT(*) FROM authorships LIMIT 1')
     Authorship.order(:name).map(:name).should == ['Bob', 'Jim']
-    sqls_should_be('SELECT * FROM authorships ORDER BY name')
     authorship2.name.should == 'Jim'
     authorship2.author_id.should == @author.id
     @author.authorships_dataset.find_or_create(:name=>'Jim').should == authorship2
-    sqls_should_be("SELECT * FROM authorships WHERE ((author_id = 1) AND (name = 'Jim')) LIMIT 1")
   end
 end
 
 describe "has_many :through has_many and has_one :through belongs_to" do
   before do
-    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.instance_variable_set(:@schemas, {})
     INTEGRATION_DB.create_table!(:firms) do
       primary_key :id
     end
@@ -291,9 +254,7 @@ describe "has_many :through has_many and has_one :through belongs_to" do
     clear_sqls
   end
   after do
-    Invoice.drop_table
-    Client.drop_table
-    Firm.drop_table
+    INTEGRATION_DB.drop_table :invoices, :clients, :firms
     Object.send(:remove_const, :Firm)
     Object.send(:remove_const, :Client)
     Object.send(:remove_const, :Invoice)
@@ -301,19 +262,16 @@ describe "has_many :through has_many and has_one :through belongs_to" do
 
   it "should return has_many :through has_many records for a single object" do
     invs = @firm1.invoices.sort_by{|x| x.pk}
-    sqls_should_be("SELECT invoices.id, invoices.client_id, client.id AS 'client_id_0', client.firm_id FROM invoices LEFT OUTER JOIN clients AS 'client' ON (client.id = invoices.client_id) WHERE (client.firm_id = 1)")
     invs.should == [@invoice1, @invoice2, @invoice3]
     invs[0].client.should == @client1
     invs[1].client.should == @client1
     invs[2].client.should == @client2
     invs.collect{|i| i.firm}.should == [@firm1, @firm1, @firm1]
     invs.collect{|i| i.client.firm}.should == [@firm1, @firm1, @firm1]
-    sqls_should_be
   end
 
   it "should eagerly load has_many :through has_many records for multiple objects" do
     firms = Firm.order(:id).eager(:invoices).all
-    sqls_should_be("SELECT * FROM firms ORDER BY id", "SELECT invoices.id, invoices.client_id, client.id AS 'client_id_0', client.firm_id FROM invoices LEFT OUTER JOIN clients AS 'client' ON (client.id = invoices.client_id) WHERE (client.firm_id IN (1, 2))")
     firms.should == [@firm1, @firm2]
     firm1, firm2 = firms
     invs1 = firm1.invoices.sort_by{|x| x.pk}
@@ -329,22 +287,18 @@ describe "has_many :through has_many and has_one :through belongs_to" do
     invs2.collect{|i| i.firm}.should == [@firm2, @firm2]
     invs1.collect{|i| i.client.firm}.should == [@firm1, @firm1, @firm1]
     invs2.collect{|i| i.client.firm}.should == [@firm2, @firm2]
-    sqls_should_be
   end
 
   it "should return has_one :through belongs_to records for a single object" do
     firm = @invoice1.firm
-    sqls_should_be("SELECT firms.id, clients.id AS 'clients_id', clients.firm_id FROM firms LEFT OUTER JOIN clients ON (clients.firm_id = firms.id) WHERE (clients.id = 1)")
     firm.should == @firm1
     @invoice1.client.should == @client1
     @invoice1.client.firm.should == @firm1
-    sqls_should_be
     firm.associations[:clients].should == nil
   end
 
   it "should eagerly load has_one :through belongs_to records for multiple objects" do
     invs = Invoice.order(:id).eager(:firm).all
-    sqls_should_be("SELECT * FROM invoices ORDER BY id", "SELECT firms.id, clients.id AS 'clients_id', clients.firm_id FROM firms LEFT OUTER JOIN clients ON (clients.firm_id = firms.id) WHERE (clients.id IN (1, 2, 3))")
     invs.should == [@invoice1, @invoice2, @invoice3, @invoice4, @invoice5]
     invs[0].firm.should == @firm1
     invs[0].client.should == @client1
@@ -366,22 +320,22 @@ describe "has_many :through has_many and has_one :through belongs_to" do
     invs[4].client.should == @client3
     invs[4].client.firm.should == @firm2
     invs[4].firm.associations[:clients].should == nil
-    sqls_should_be
   end
 end
 
 describe "Polymorphic Associations" do
   before do
-    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.instance_variable_set(:@schemas, {})
     INTEGRATION_DB.create_table!(:assets) do
       primary_key :id
-      integer :attachable_id
-      text :attachable_type
+      Integer :attachable_id
+      String :attachable_type
     end
     class ::Asset < Sequel::Model
+      m = method(:constantize)
       many_to_one :attachable, :reciprocal=>:assets, \
         :dataset=>(proc do
-          klass = attachable_type.constantize
+          klass = m.call(attachable_type)
           klass.filter(klass.primary_key=>attachable_id)
         end), \
         :eager_loader=>(proc do |key_hash, assets, associations|
@@ -391,7 +345,7 @@ describe "Polymorphic Associations" do
             ((id_map[asset.attachable_type] ||= {})[asset.attachable_id] ||= []) << asset
           end 
           id_map.each do |klass_name, id_map|
-            klass = klass_name.constantize
+            klass = m.call(klass_name)
             klass.filter(klass.primary_key=>id_map.keys).all do |attach|
               id_map[attach.pk].each do |asset|
                 asset.associations[:attachable] = attach
@@ -412,7 +366,7 @@ describe "Polymorphic Associations" do
       primary_key :id
     end
     class ::Post < Sequel::Model
-      one_to_many :assets, :key=>:attachable_id do |ds|
+      one_to_many :assets, :key=>:attachable_id, :reciprocal=>:attachable do |ds|
         ds.filter(:attachable_type=>'Post')
       end 
       
@@ -438,7 +392,7 @@ describe "Polymorphic Associations" do
       primary_key :id
     end
     class ::Note < Sequel::Model
-      one_to_many :assets, :key=>:attachable_id do |ds|
+      one_to_many :assets, :key=>:attachable_id, :reciprocal=>:attachable do |ds|
         ds.filter(:attachable_type=>'Note')
       end 
       
@@ -469,9 +423,7 @@ describe "Polymorphic Associations" do
     clear_sqls
   end
   after do
-    Asset.drop_table
-    Post.drop_table
-    Note.drop_table
+    INTEGRATION_DB.drop_table :assets, :posts, :notes
     Object.send(:remove_const, :Asset)
     Object.send(:remove_const, :Post)
     Object.send(:remove_const, :Note)
@@ -480,22 +432,18 @@ describe "Polymorphic Associations" do
   it "should load the correct associated object for a single object" do
     @asset1.attachable.should == @post
     @asset2.attachable.should == @note
-    sqls_should_be("SELECT * FROM posts WHERE (id = 1) LIMIT 1", "SELECT * FROM notes WHERE (id = 2) LIMIT 1")
   end
 
   it "should eagerly load the correct associated object for a group of objects" do
     assets = Asset.order(:id).eager(:attachable).all
-    sqls_should_be("SELECT * FROM assets ORDER BY id", "SELECT * FROM posts WHERE (id IN (1))", "SELECT * FROM notes WHERE (id IN (2))")
     assets.should == [@asset1, @asset2]
     assets[0].attachable.should == @post
     assets[1].attachable.should == @note
-    sqls_should_be
   end
 
   it "should set items correctly" do
     @asset1.attachable = @note
     @asset2.attachable = @post
-    sqls_should_be("SELECT * FROM posts WHERE (id = 1) LIMIT 1", "SELECT * FROM notes WHERE (id = 2) LIMIT 1")
     @asset1.attachable.should == @note
     @asset1.attachable_id.should == @note.pk
     @asset1.attachable_type.should == 'Note'
@@ -506,38 +454,29 @@ describe "Polymorphic Associations" do
     @asset1.attachable.should == nil
     @asset1.attachable_id.should == nil
     @asset1.attachable_type.should == nil
-    sqls_should_be
   end
 
   it "should add items correctly" do
     @post.assets.should == [@asset1]
-    sqls_should_be("SELECT * FROM assets WHERE ((assets.attachable_id = 1) AND (attachable_type = 'Post'))")
     @post.add_asset(@asset2)
-    sqls_should_be(/UPDATE assets SET ((attachable_id = 1|attachable_type = 'Post'|id = 2)(, )?){3} WHERE \(id = 2\)/)
     @post.assets.should == [@asset1, @asset2]
     @asset2.attachable.should == @post
     @asset2.attachable_id.should == @post.pk
     @asset2.attachable_type.should == 'Post'
-    sqls_should_be
   end
 
   it "should remove items correctly" do
     @note.assets.should == [@asset2]
-    sqls_should_be("SELECT * FROM assets WHERE ((assets.attachable_id = 2) AND (attachable_type = 'Note'))")
     @note.remove_asset(@asset2)
-    sqls_should_be(/UPDATE assets SET ((attachable_id = NULL|attachable_type = NULL|id = 2)(, )?){3} WHERE \(id = 2\)/)
     @note.assets.should == []
     @asset2.attachable.should == nil
     @asset2.attachable_id.should == nil
     @asset2.attachable_type.should == nil
-    sqls_should_be
   end
 
   it "should remove all items correctly" do
     @post.remove_all_assets
     @note.remove_all_assets
-    sqls_should_be(/UPDATE assets SET attachable_(id|type) = NULL, attachable_(type|id) = NULL WHERE \(\(attachable_(id|type) = (1|'Post')\) AND \(attachable_(type|id) = ('Post'|1)\)\)/,
-      /UPDATE assets SET attachable_(id|type) = NULL, attachable_(type|id) = NULL WHERE \(\(attachable_(id|type) = (2|'Note')\) AND \(attachable_(type|id) = ('Note'|2)\)\)/)
     @asset1.reload.attachable.should == nil
     @asset2.reload.attachable.should == nil
   end
@@ -545,10 +484,10 @@ end
 
 describe "many_to_one/one_to_many not referencing primary key" do
   before do
-    INTEGRATION_DB.instance_variable_set(:@schemas, nil)
+    INTEGRATION_DB.instance_variable_set(:@schemas, {})
     INTEGRATION_DB.create_table!(:clients) do
       primary_key :id
-      text :name
+      String :name
     end
     class ::Client < Sequel::Model
       one_to_many :invoices, :reciprocal=>:client, \
@@ -582,7 +521,7 @@ describe "many_to_one/one_to_many not referencing primary key" do
   
     INTEGRATION_DB.create_table!(:invoices) do
       primary_key :id
-      text :client_name
+      String :client_name
     end
     class ::Invoice < Sequel::Model
       many_to_one :client, :key=>:client_name, \
@@ -609,88 +548,132 @@ describe "many_to_one/one_to_many not referencing primary key" do
     clear_sqls
   end
   after do
-    Invoice.drop_table
-    Client.drop_table
+    INTEGRATION_DB.drop_table :invoices, :clients
     Object.send(:remove_const, :Client)
     Object.send(:remove_const, :Invoice)
   end
 
   it "should load all associated one_to_many objects for a single object" do
     invs = @client1.invoices
-    sqls_should_be("SELECT * FROM invoices WHERE (client_name = 'X')")
     invs.should == [@invoice1, @invoice2]
     invs[0].client.should == @client1
     invs[1].client.should == @client1
-    sqls_should_be
   end
 
   it "should load the associated many_to_one object for a single object" do
     client = @invoice1.client
-    sqls_should_be("SELECT * FROM clients WHERE (name = 'X') LIMIT 1")
     client.should == @client1
   end
 
   it "should eagerly load all associated one_to_many objects for a group of objects" do
     clients = Client.order(:id).eager(:invoices).all
-    sqls_should_be("SELECT * FROM clients ORDER BY id", "SELECT * FROM invoices WHERE (client_name IN ('X', 'Y'))")
     clients.should == [@client1, @client2]
     clients[1].invoices.should == []
     invs = clients[0].invoices.sort_by{|x| x.pk}
     invs.should == [@invoice1, @invoice2]
     invs[0].client.should == @client1
     invs[1].client.should == @client1
-    sqls_should_be
   end
 
   it "should eagerly load the associated many_to_one object for a group of objects" do
     invoices = Invoice.order(:id).eager(:client).all
-    sqls_should_be("SELECT * FROM invoices ORDER BY id", "SELECT * FROM clients WHERE (name IN ('X'))")
     invoices.should == [@invoice1, @invoice2]
     invoices[0].client.should == @client1
     invoices[1].client.should == @client1
-    sqls_should_be
   end
 
   it "should set the associated object correctly" do
     @invoice1.client = @client2
-    sqls_should_be("SELECT * FROM clients WHERE (name = 'X') LIMIT 1")
     @invoice1.client.should == @client2
     @invoice1.client_name.should == 'Y'
     @invoice1.client = nil
     @invoice1.client_name.should == nil
-    sqls_should_be
   end
 
   it "should add the associated object correctly" do
     @client2.invoices.should == []
-    sqls_should_be("SELECT * FROM invoices WHERE (client_name = 'Y')")
     @client2.add_invoice(@invoice1)
-    sqls_should_be(/UPDATE invoices SET (client_name = 'Y'|id = 1), (client_name = 'Y'|id = 1) WHERE \(id = 1\)/)
     @client2.invoices.should == [@invoice1]
     @invoice1.client_name.should == 'Y'
     @invoice1.client = nil
     @invoice1.client_name.should == nil
-    sqls_should_be
   end
 
   it "should remove the associated object correctly" do
     invs = @client1.invoices.sort_by{|x| x.pk}
-    sqls_should_be("SELECT * FROM invoices WHERE (client_name = 'X')")
     invs.should == [@invoice1, @invoice2]
     @client1.remove_invoice(@invoice1)
-    sqls_should_be(/UPDATE invoices SET (client_name = NULL|id = 1), (client_name = NULL|id = 1) WHERE \(id = 1\)/)
     @client1.invoices.should == [@invoice2]
     @invoice1.client_name.should == nil
     @invoice1.client.should == nil
-    sqls_should_be
   end
 
   it "should remove all associated objects correctly" do
     invs = @client1.remove_all_invoices
-    sqls_should_be("UPDATE invoices SET client_name = NULL WHERE (client_name = 'X')")
     @invoice1.refresh.client.should == nil
     @invoice1.client_name.should == nil
     @invoice2.refresh.client.should == nil
     @invoice2.client_name.should == nil
   end
 end
+
+describe "statistics associations" do
+  before do
+    INTEGRATION_DB.create_table!(:projects) do
+      primary_key :id
+      String :name
+    end
+    class ::Project < Sequel::Model
+      many_to_one :ticket_hours, :read_only=>true, :key=>:id,
+       :dataset=>proc{Ticket.filter(:project_id=>id).select{sum(hours).as(hours)}},
+       :eager_loader=>(proc do |kh, projects, a|
+        projects.each{|p| p.associations[:ticket_hours] = nil}
+        Ticket.filter(:project_id=>kh[:id].keys).
+         group(:project_id).
+         select{[project_id.as(project_id), sum(hours).as(hours)]}.
+         all do |t|
+          p = kh[:id][t.values.delete(:project_id)].first
+          p.associations[:ticket_hours] = t
+         end
+       end)  
+      def ticket_hours
+        if s = super
+          s[:hours]
+        end
+      end 
+    end 
+
+    INTEGRATION_DB.create_table!(:tickets) do
+      primary_key :id
+      foreign_key :project_id, :projects
+      Integer :hours
+    end
+    class ::Ticket < Sequel::Model
+      many_to_one :project
+    end
+
+    @project1 = Project.create(:name=>'X')
+    @project2 = Project.create(:name=>'Y')
+    @ticket1 = Ticket.create(:project=>@project1, :hours=>1)
+    @ticket2 = Ticket.create(:project=>@project1, :hours=>10)
+    @ticket3 = Ticket.create(:project=>@project2, :hours=>2)
+    @ticket4 = Ticket.create(:project=>@project2, :hours=>20)
+    clear_sqls
+  end
+  after do
+    INTEGRATION_DB.drop_table :tickets, :projects
+    Object.send(:remove_const, :Project)
+    Object.send(:remove_const, :Ticket)
+  end
+
+  it "should give the correct sum of ticket hours for each project" do
+    @project1.ticket_hours.to_i.should == 11
+    @project2.ticket_hours.to_i.should == 22
+  end
+
+  it "should give the correct sum of ticket hours for each project when eager loading" do
+    p1, p2 = Project.order(:name).eager(:ticket_hours).all
+    p1.ticket_hours.to_i.should == 11
+    p2.ticket_hours.to_i.should == 22
+  end
+end
